#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "lcg.h"
#include "bitmap.h"

// ----------------------------------------------------------------
struct cmd_line_opts_t
{
	int    mxdim;
	int    num_reps;
	int    num_states;
	double e;
	int    chained;

	int    show_states;

	int    have_specified_seeds;
	int    seed_1;
	int    seed_2;
	int    non_random_initial_placement;
	int    zero_initial_placement;

	int    use_binary_input_file;
	char * binary_input_file_name;

	int    want_binary_output_file;
	char * binary_output_file_name;

	int    want_matrix_bitmap_file;
	char * matrix_bitmap_file_name;

	int    want_state_bitmap_file;
	char * state_bitmap_file_name;

	int    have_trace_cell;
	int    trace_row;
	int    trace_column;
	int    dump_when_did_hebb;
	int    trace_trajectories;
	int    dump_final_matrix_elements;
	double dump_final_matrix_element_threshold;

	int    print_first_matrix;
	int    print_last_matrix;

	int    want_q;
	double q;
};

// ----------------------------------------------------------------
static int parse_command_line(
	int      argc,
	char  ** argv,
	struct cmd_line_opts_t * popts);

static void usage(
	char * argv0);

void seed_random_number_generator(
	int have_specified_seeds,
	int seed_1,
	int seed_2);

double random_unit_double(
	void);

void hmc(
	struct cmd_line_opts_t * popts);

double ** alloc_matrix(
	int mxdim);

void init_matrix(
	double ** P,
	int       mxdim);

void init_matrix_flat(
	double ** P,
	int       mxdim);

void init_matrix_block(
	double ** P,
	int       mxdim);

void init_matrix_holes(
	double ** P,
	int       mxdim);

double get_row_sum(
	double ** P,
	int       mxdim,
	int       row_index);

int get_row_peak_index(
	double ** P,
	int       mxdim,
	int       row_index);

void normalize_row(
	double ** P,
	int       mxdim,
	int       row_index);

void show_matrix_with_row_sums(
	double ** P,
	int       mxdim);

int hmc_next_state(
	double ** P,
	int       mxdim,
	int       current_state);

int apply_hebb_rule(
	double ** P,
	int       mxdim,
	struct cmd_line_opts_t * popts,
	int     * current_states,
	int     * next_states,
	int       num_states);

void dump_matrix_elements(
	double ** P,
	int       mxdim,
	double    threshold);

void load_matrix_from_binary_file(
	double ** P,
	int       mxdim,
	int     * current_states,
	int       num_states,
	char    * fn);

void store_matrix_to_binary_file(
	double ** P,
	int       mxdim,
	int     * current_states,
	int       num_states,
	char    * fn);

void store_matrix_to_bitmap_file(
	double ** P,
	int       mxdim,
	char    * fn,
	int       pixels_per_cell);

void store_states_to_bitmap_file(
	int     * current_states,
	int     * next_states,
	int       num_states,
	int       mxdim,
	char    * fn,
	int       pixels_per_cell);

void free_matrix(
	double ** P,
	int       mxdim);

// ================================================================
int main(int argc, char **argv)
{
	struct cmd_line_opts_t cmd_line_opts;

	if (!parse_command_line(argc, argv, &cmd_line_opts))
		usage(argv[0]);

	seed_random_number_generator(
		cmd_line_opts.have_specified_seeds,
		cmd_line_opts.seed_1,
		cmd_line_opts.seed_2);

	hmc(&cmd_line_opts);

	return 0;
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [options] {matrix dim} {# reps}\n", argv0);

	fprintf(stderr,
	"Options:\n"
	"  /s {s1}:{s2}      : Provide specific seeds for random-number generator.\n"
	"                      If omitted, time of day is used.  Numbers are assumed\n"
	"                      to be in hex; do not use 0x prefix.\n"
	"  /m {# states}     : Default matrix dimension if omitted.\n"
	"  /e {coupling}     : Default 0.1.\n"
	"  /u                : Unchained.  Default is chained.\n"
	"  /o {binary file}  : Save matrix and states in binary file.\n"
	"  /i {binary file}  : Resume from binary file.\n"
	"  /bm {bitmap file} : Bitmap file for matrix.\n"
	"  /bt {bitmap file} : Bitmap file for most recent transitions.\n"
	"  /c {x, y}         : Trace transitions at specified cell.  Currently inoperative.\n"
	"  /f                : Print first matrix to screen.\n"
	"  /l                : Print last  matrix to screen.\n"
	"  /st               : Print state transitions to screen, for last 200 iterations.\n"
	"\n");

	exit(1);
}

// ----------------------------------------------------------------
static int parse_command_line(
	int      argc,
	char  ** argv,
	struct cmd_line_opts_t * popts)
{
	int argi;
	int got_num_states = 0;

 	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Apply defaults.
	popts->mxdim                   = 1;
	popts->num_reps                = 0;
	popts->num_states              = 2;
	popts->e                       = 0.1;
	popts->chained                 = 1;

	popts->show_states             = 0;

	popts->have_specified_seeds    = 0;
	popts->seed_1                  = 0;
	popts->seed_2                  = 0;
	popts->non_random_initial_placement = 0;
	popts->zero_initial_placement  = 0;

	popts->use_binary_input_file   = 0;
	popts->binary_input_file_name  = 0;

	popts->want_binary_output_file = 0;
	popts->binary_output_file_name = 0;

	popts->want_matrix_bitmap_file = 0;
	popts->matrix_bitmap_file_name = 0;

	popts->want_state_bitmap_file  = 0;
	popts->state_bitmap_file_name  = 0;

	popts->have_trace_cell         = 0;
	popts->trace_row               = 0;
	popts->trace_column            = 0;
	popts->dump_when_did_hebb      = 0;
	popts->trace_trajectories      = 0;
	popts->dump_final_matrix_elements = 0;
	popts->dump_final_matrix_element_threshold = 0;

	popts->print_first_matrix      = 0;
	popts->print_last_matrix       = 0;

	popts->want_q                  = 0;
	popts->q                       = 0.0;

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	for (argi = 1; argi < argc; /* variable increment, done inside loop */ ) {

		if (argv[argi][0] != '/') {
			break;
		}

		else if (strcmp(&argv[argi][1], "s") == 0) {
			if (sscanf(argv[argi+1], "%x:%x", &popts->seed_1, &popts->seed_2) != 2)
				if (sscanf(argv[argi+1], "%x,%x", &popts->seed_1, &popts->seed_2) != 2)
					return 0;
			popts->have_specified_seeds = 1;
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "m") == 0) {
			if (sscanf(argv[argi+1], "%d", &popts->num_states) != 1)
				return 0;
			if (popts->num_states < 2)
				return 0;
			got_num_states = 1;
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "e") == 0) {
			if (sscanf(argv[argi+1], "%lf", &popts->e) != 1)
				return 0;
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "q") == 0) {
			if (sscanf(argv[argi+1], "%lf", &popts->q) != 1)
				return 0;
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "u") == 0) {
			popts->chained = 0;
			argi += 1;
		}

		else if (strcmp(&argv[argi][1], "nrip") == 0) {
			popts->non_random_initial_placement = 1;
			argi += 1;
		}

		else if (strcmp(&argv[argi][1], "zip") == 0) {
			popts->zero_initial_placement = 1;
			argi += 1;
		}

		else if (strcmp(&argv[argi][1], "j") == 0) {
			popts->trace_trajectories = 1;
			argi += 1;
		}

		else if (strcmp(&argv[argi][1], "dm") == 0) {
			if (sscanf(argv[argi+1], "%lf",
					&popts->dump_final_matrix_element_threshold) != 1)
				return 0;
			popts->dump_final_matrix_elements = 1;
			argi += 2;
		}


		else if (strcmp(&argv[argi][1], "i") == 0) {
			popts->use_binary_input_file = 1;
			popts->binary_input_file_name = argv[argi+1];
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "o") == 0) {
			popts->want_binary_output_file = 1;
			popts->binary_output_file_name = argv[argi+1];
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "bm") == 0) {
			popts->want_matrix_bitmap_file = 1;
			popts->matrix_bitmap_file_name = argv[argi+1];
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "bt") == 0) {
			popts->want_state_bitmap_file = 1;
			popts->state_bitmap_file_name = argv[argi+1];
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "c") == 0) {
			if (sscanf(argv[argi+1], "%d,%d", &popts->trace_row, &popts->trace_column) != 2)
				return 0;
			popts->have_trace_cell = 1;
			argi += 2;
		}

		else if (strcmp(&argv[argi][1], "dh") == 0) {
			popts->dump_when_did_hebb = 1;
			argi += 1;
		}

		else if (strcmp(&argv[argi][1], "f") == 0) {
			popts->print_first_matrix = 1;
			argi += 1;
		}

		else if (strcmp(&argv[argi][1], "l") == 0) {
			popts->print_last_matrix = 1;
			argi += 1;
		}

		else if (strcmp(&argv[argi][1], "st") == 0) {
			popts->show_states = 1;
			argi += 1;
		}

		else
			return 0;
	}

	if (argc - argi != 2)
		return 0;

	if (sscanf(argv[argi],     "%d", &popts->mxdim) != 1)
		return 0;
	if (sscanf(argv[argi + 1], "%d", &popts->num_reps) != 1)
		return 0;

	if (!got_num_states)
		popts->num_states = popts->mxdim;

	return 1;
}

// ----------------------------------------------------------------
void hmc(
	struct cmd_line_opts_t * popts)
{
	int mxdim = popts->mxdim;
	int num_reps = popts->num_reps;
	int i;
	int j;
	int * current_states;
	int * next_states;
	double ** P;
	int pixels_per_cell;
	int min_bitmap_size = 300;
	int did_hebb;

	P = alloc_matrix(mxdim);
	current_states = (int *) malloc (popts->num_states * sizeof(int));
	next_states    = (int *) malloc (popts->num_states * sizeof(int));
	if ((current_states == 0) || (next_states == 0)) {
		fprintf(stderr, "malloc(%d) failed.\n",
			popts->num_states * sizeof(int));
		exit(1);
	}

	if (popts->use_binary_input_file) {
		load_matrix_from_binary_file(P, mxdim, current_states, popts->num_states,
			popts->binary_input_file_name);
	}
	else {
		// init_matrix(P, mxdim);
		init_matrix_flat(P, mxdim);
		// init_matrix_holes(P, mxdim);
		// init_matrix_block(P, mxdim);

		for (i = 0; i < popts->num_states; i++) {
			if (popts->zero_initial_placement)
				current_states[i] = 0;
			else if (popts->non_random_initial_placement)
				current_states[i] = (i % mxdim);
			else
				current_states[i] = (int) (random_unit_double() * mxdim);
			next_states[i]    = 0;
		}
	}

	if (popts->print_first_matrix) {
		printf("Initial matrix:\n");
		show_matrix_with_row_sums(P, mxdim);
	}

	for (i = 0; i < num_reps; i++) {

		// printf("=== Repetition %d ===========================\n", i);

		if (popts->trace_trajectories) {
			int k;
			printf("%d", i);
			for (k = 0; k < popts->num_states; k++) {
				printf("\t%d", current_states[k]);
			}
			printf("\n");
		}

		for (j = 0; j < popts->num_states; j++)
			next_states[j] = hmc_next_state(P, mxdim, current_states[j]);

		did_hebb = apply_hebb_rule(P, mxdim, popts, 
			current_states, next_states, popts->num_states);

		//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		if (did_hebb && popts->dump_when_did_hebb) {
			printf("Modified at step %d.\n", i);
			show_matrix_with_row_sums(P, mxdim);
		}

		if (popts->show_states) {
			if (i > (num_reps - 200)) {
				printf("Rep %7d:", i);
				for (j = 0; j < popts->num_states; j++)
					printf(" %3d->%3d",
					current_states[j], next_states[j]);
				printf("\n");
			}
		}

		//	if (popts->have_trace_cell) {
		//		if (did_hebb) {
		//			if (((popts->trace_row == A) && (popts->trace_column == A_next))
		//			||  ((popts->trace_row == B) && (popts->trace_column == B_next))) {
		//				printf("Rep %7d: (%3d, %3d) = %7.4g (H: %3d, %3d -> %3d)\n",
		//					i,
		//					popts->trace_row,
		//					popts->trace_column,
		//					P[popts->trace_row][popts->trace_column],
		//					A,
		//					B,
		//					A_next);
		//			}
		//		}
		//	}


		//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		if (i < (num_reps - 1)) {
			if (popts->chained) {
				for (j = 0; j < popts->num_states; j++)
					current_states[j] = next_states[j];
			}
			else { // unchained
				for (j = 0; j < popts->num_states; j++)
					current_states[j] = (int) (random_unit_double() * mxdim);
			}
		}

		// printf("\n");

		if (num_reps > 20) {
			if (!popts->dump_when_did_hebb && !popts->trace_trajectories) {
				if ((i % (num_reps / 10)) == 0) {
					printf("%d%% ", (int) (100.0 * (double)i / (double)num_reps));
				}
			}
		}
	}
	if (num_reps > 20)
		if (!popts->dump_when_did_hebb && !popts->trace_trajectories)
			printf("100%%\n");


	if (popts->print_last_matrix) {
		printf("Final matrix:\n");
		show_matrix_with_row_sums(P, mxdim);
	}



	if (mxdim > min_bitmap_size)
		pixels_per_cell = 1;
	else
		pixels_per_cell = min_bitmap_size / mxdim;

	if (popts->want_matrix_bitmap_file) {
		printf("Storing matrix to bitmap file \"%s\" ... ", popts->matrix_bitmap_file_name);
		fflush(stdout);
		store_matrix_to_bitmap_file(P, mxdim, popts->matrix_bitmap_file_name,
			pixels_per_cell);
		printf("done.\n");
	}

	if (popts->want_state_bitmap_file) {
		printf("Storing states to bitmap file \"%s\" ... ", popts->state_bitmap_file_name);
		fflush(stdout);
		store_states_to_bitmap_file(current_states, next_states, popts->num_states,
			popts->mxdim, popts->state_bitmap_file_name, pixels_per_cell);
		printf("done.\n");
	}

	if (popts->want_binary_output_file)
		store_matrix_to_binary_file(P, mxdim, current_states, popts->num_states,
			popts->binary_output_file_name);

	if (popts->dump_final_matrix_elements)
		dump_matrix_elements(P, mxdim, popts->dump_final_matrix_element_threshold);

	free_matrix(P, mxdim);
	free(current_states);
	free(next_states);
}

// ----------------------------------------------------------------
void seed_random_number_generator(
	int have_specified_seeds,
	int seed_1,
	int seed_2)
{
	// unsigned srand_seed;
	unsigned LCG_seed_1;
	unsigned LCG_seed_2;

	if (!have_specified_seeds) {
		SYSTEMTIME st;
		FILETIME   ft;

		GetSystemTime(&st);
		SystemTimeToFileTime(&st, &ft);
		seed_1 = ft.dwLowDateTime;
		seed_2 = ft.dwHighDateTime;
	}

	printf("General seeds = 0x%08x 0x%08x\n", seed_1, seed_2);

	// srand_seed = seed_1 ^ seed_2;
	// srand(srand_seed);

	LCG_seed_1 = seed_1 ^ seed_2;
	LCG_seed_2 = seed_1 * seed_1;
	initLCG(LCG_seed_1, LCG_seed_2);
	printf("Seeds for LCG = 0x%08x 0x%08x\n", LCG_seed_1, LCG_seed_2);
}

// ----------------------------------------------------------------
double random_unit_double(
	void)
{
	// return (double)rand() / (double)RAND_MAX;
	return combinedLCG();
}

// ----------------------------------------------------------------
double ** alloc_matrix(
	int mxdim)
{
	double ** P;
	int nbytes;
	int i;

	nbytes = mxdim * sizeof(double *);
	P = (double **) malloc(nbytes);

	if (P == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", nbytes);
		exit(1);
	}

	for (i = 0; i < mxdim; i++) {
		nbytes = mxdim * sizeof(double);
		P[i] = (double *) malloc(nbytes);

		if (P[i] == 0) {
			fprintf(stderr, "malloc(%d) failed.\n", nbytes);
			exit(1);
		}
	}

	return P;
}

// ----------------------------------------------------------------
void init_matrix(
	double ** P,
	int mxdim)
{
	int i;
	int j;

	for (i = 0; i < mxdim; i++) {
		for (j = 0; j < mxdim; j++) {
			P[i][j] = random_unit_double();
		}
		normalize_row(P, mxdim, i);
	}
}

// ----------------------------------------------------------------
void init_matrix_flat(
	double ** P,
	int mxdim)
{
	int i;
	int j;

	for (i = 0; i < mxdim; i++) {
		for (j = 0; j < mxdim; j++) {
			P[i][j] = 1.0 / mxdim;
		}
		normalize_row(P, mxdim, i);
	}
}

// ----------------------------------------------------------------
void init_matrix_block(
	double ** P,
	int mxdim)
{
	int i;
	int j;
	int num_blocks = 8;
	int block_len  = mxdim / num_blocks;
	int iblockno;
	int jblockno;
	int istart;
	int iend;
	int jstart;
	int jend;

	for (i = 0; i < mxdim; i++) {
		for (j = 0; j < mxdim; j++) {
			iblockno = i / block_len;
			jblockno = j / block_len;
			if (iblockno == jblockno)
				P[i][j] = random_unit_double();
			else
				P[i][j] = 0.0;
		}
		normalize_row(P, mxdim, i);
	}
}

// ----------------------------------------------------------------
void init_matrix_holes(
	double ** P,
	int mxdim)
{
	int i;
	int j;
	double zero_fraction = 0.9;

	for (i = 0; i < mxdim; i++) {
		for (j = 0; j < mxdim; j++) {
			if (random_unit_double() < zero_fraction)
				P[i][j] = 0.0;
			else
				P[i][j] = random_unit_double();
		}
		normalize_row(P, mxdim, i);
	}
}

// ----------------------------------------------------------------
double get_row_sum(
	double ** P,
	int       mxdim,
	int       row_index)
{
	double row_sum = 0.0;
	int j;

	for (j = 0; j < mxdim; j++)
		row_sum += P[row_index][j];

	return row_sum;
}

// ----------------------------------------------------------------
int get_row_peak_index(
	double ** P,
	int       mxdim,
	int       row_index)
{
	int    row_peak_index = 0;
	double row_peak_value = 0.0;
	int j;

	for (j = 0; j < mxdim; j++) {
		if (P[row_index][j] > row_peak_value) {
			row_peak_value = P[row_index][j];
			row_peak_index = j;
		}
	}

	return row_peak_index;
}

// ----------------------------------------------------------------
void normalize_row(
	double ** P,
	int       mxdim,
	int       row_index)
{
	double sum = get_row_sum(P, mxdim, row_index);
	int j;

	if (sum != 0.0) {
		double sum_recip = 1.0 / sum;
		for (j = 0; j < mxdim; j++)
			P[row_index][j] *= sum_recip;
	}
}

// ----------------------------------------------------------------
void show_matrix_with_row_sums(
	double ** P,
	int mxdim)
{
	int i;
	int j;
	int k;
	double row_sum;
	int row_peak_index;
	int field_width        = 7;
	int num_decimal_places = 4;
	char printf_format_non_peak[20];
	char printf_format_peak[20];

	sprintf(printf_format_non_peak, "%%%d.%df ",
		field_width, num_decimal_places);
	sprintf(printf_format_peak, "%%%d.%df<",
		field_width, num_decimal_places);

	for (i = 0; i < mxdim; i++) {
		row_sum        = get_row_sum(P, mxdim, i);
		row_peak_index = get_row_peak_index(P, mxdim, i);

		for (j = 0; j < mxdim; j++) {
			if (j == row_peak_index)
				printf(printf_format_peak, P[i][j]);
			else
				printf(printf_format_non_peak, P[i][j]);
			if (j < mxdim - 1)
				printf(" ");
		}
		printf(" | ");
		printf(printf_format_non_peak, row_sum);
		printf("\n");
	}

	printf("\n");
}

// ----------------------------------------------------------------
int hmc_next_state(
	double ** P,
	int       mxdim,
	int       current_state)
{
	double left_sum;
	double right_sum;
	double selector;
	int next_state = mxdim - 1;
	int j;

	selector  = random_unit_double();

	for (j = 0; j < mxdim; j++) {
		if (j == 0) {
			left_sum   = 0.0;
			right_sum  = P[current_state][0];
		}
		else {
			left_sum  += P[current_state][j-1];
			right_sum += P[current_state][j];
		}

		if ((left_sum <= selector) && (selector < right_sum)) {
			return j;
		}
	}

	printf("Ouch!  Selector %7.4f; left %7.4f; right %7.4f; row sum %7.4f\n",
		selector, left_sum, right_sum, get_row_sum(P, mxdim, current_state));
	exit(1);

	return next_state;
}

// ----------------------------------------------------------------
int apply_hebb_rule(
	double ** P,
	int       mxdim,
	struct cmd_line_opts_t * popts,
	int     * current_states,
	int     * next_states,
	int       num_states)
{
	double Pa;
	double Pb;
	int A;
	int B;
	int A_next;
	int B_next;
	int i;
	int j;
	int num_changes = 0;

	for (i = 0; i < num_states; i++) {
		for (j = 0; j < i; j++) {

			A      = current_states[i];
			B      = current_states[j];
			A_next = next_states[i];
			B_next = next_states[j];

			// Hebb's rule:
			if (A_next != B_next)
				continue;

			// Don't do anything for self-stimulation:
			if (A == B)
				continue;
			if (A == A_next)
				continue;
			if (B == B_next)
				continue;

			if (popts->dump_when_did_hebb) {
				printf("Applied Hebb rule at %d,%d and %d,%d.\n",
					A, A_next, B, B_next);
			}
			Pa = P[A][A_next];
			Pb = P[B][B_next];

			if (popts->want_q) {
				P[A][A_next] = popts->q;
				P[B][B_next] = popts->q;
			}
			else {
				P[A][A_next] += popts->e * Pb;
				P[B][B_next] += popts->e * Pa;

				if (P[A][A_next] < 0.0)
					P[A][A_next] = 0.0;
				if (P[B][B_next] < 0.0)
					P[B][B_next] = 0.0;
			}

			normalize_row(P, mxdim, A);
			normalize_row(P, mxdim, B);

			num_changes++;

		}
	}

	return num_changes;
}

// ----------------------------------------------------------------
void free_matrix(
	double ** P,
	int       mxdim)
{
	int i;

	for (i = 0; i < mxdim; i++)
		free(P[i]);

	free(P);
}

// ----------------------------------------------------------------
void dump_matrix_elements(
	double ** P,
	int       mxdim,
	double    threshold)
{
	int i, j;

	printf("Matrix elements exceeding %11.7lf:\n", threshold);
		for (j = 0; j < mxdim; j++)
	for (i = 0; i < mxdim; i++)
			if (P[i][j] > threshold)
				printf("P[%3d][%3d] = %11.7lf\n", i, j, P[i][j]);
	printf("\n");
}

// ----------------------------------------------------------------
void load_matrix_from_binary_file(
	double ** P,
	int       mxdim,
	int     * current_states,
	int       num_states,
	char    * fn)
{
	int i;
	FILE * fp;

	fp = fopen(fn, "rb");
	if (fp == NULL) {
		fprintf(stderr,
			"Couldn't open binary input file \"%s\" for read.  Exiting.\n",
			fn);
		exit(1);
	}

	for (i = 0; i < mxdim; i++) {
		if (fread(&P[i][0], sizeof(P[i][0]), mxdim, fp) != mxdim) {
			fprintf(stderr,
				"Failed fread() from binary input file \"%s\" on row %d.  Exiting.\n",
				fn, i);
			exit(1);
		}
	}

	for (i = 0; i < num_states; i++) {
		if (fread(&current_states[i], sizeof(current_states[i]), 1, fp) != 1) {
			fprintf(stderr,
			"Failed fread() from binary input file \"%s\" on state %d.  Exiting.\n",
				fn, i);
				exit(1);
		}
	}

	fclose(fp);
}

// ----------------------------------------------------------------
void store_matrix_to_binary_file(
	double ** P,
	int       mxdim,
	int     * current_states,
	int       num_states,
	char    * fn)
{
	int i;
	FILE * fp;

	fp = fopen(fn, "wb");
	if (fp == NULL) {
		fprintf(stderr,
			"Couldn't open binary output file \"%s\" for write.  Exiting.\n",
			fn);
		exit(1);
	}

	for (i = 0; i < mxdim; i++) {
		if (fwrite(&P[i][0], sizeof(P[i][0]), mxdim, fp) != mxdim) {
			fprintf(stderr,
				"Failed fwrite() to binary output file \"%s\" on row %d.  Exiting.\n",
				fn, i);
			exit(1);
		}
	}

	for (i = 0; i < num_states; i++) {
		if (fwrite(&current_states[i], sizeof(current_states[i]), 1, fp) != 1) {
			fprintf(stderr,
			"Failed fwrite() to binary output file \"%s\" on state %d.  Exiting.\n",
				fn, i);
				exit(1);
		}
	}

	fclose(fp);
}

// ----------------------------------------------------------------
void store_matrix_to_bitmap_file(
	double ** P,
	int       mxdim,
	char    * fn,
	int       pixels_per_cell)
{
	struct tBitmapHeader header;
	int i;
	int j;
	int ii;
	int jj;
	FILE * fp;
	unsigned char pix[3];
	unsigned char pad[4] = {0, 0, 0, 0};
	int remaining_bytes;
	unsigned greyscale;
	double dmin;
	double dmax;
	double drange;
	double drangerecip;

	init_bitmap_header(mxdim * pixels_per_cell, mxdim * pixels_per_cell, &header);

	fp = fopen(fn, "wb");
	if (fp == 0) {
		fprintf(stderr, "Couldn't open \"%s\" for binary write. Exiting.\n", fn);
		exit(1);
	}
	fwrite(&header, sizeof(header), 1, fp);

	dmin =  P[0][0];
	dmax =  P[0][0];
	for (i = 0; i < mxdim; i++) {
		for (j = 0; j < mxdim; j++) {
			if (P[i][j] < dmin)
				dmin = P[i][j];
			if (P[i][j] > dmax)
				dmax = P[i][j];
		}
	}
	drange = dmax - dmin;
	if (drange == 0.0) {
		fprintf(stderr, "Matrix is constant %7.4f.\n", dmax);
		exit(1);
	}
	drangerecip = 1.0 / drange;
	printf("Low: %7.4f  High: %7.4f\n", dmin, dmax);

	for (i = mxdim - 1; i >= 0; i--) { // Bitmap file format is "upside down".
		for (ii = 0; ii < pixels_per_cell; ii++) {
			for (j = 0; j < mxdim; j++) {
				greyscale = (unsigned) ( ((P[i][j] - dmin) * drangerecip) * 256.0);
				if (greyscale < 0)
					greyscale = 0;
				if (greyscale >= 255)
					greyscale = 255;

				pix[0] = greyscale;
				pix[1] = greyscale;
				pix[2] = greyscale;

				if (greyscale == 0)
					pix[1] = 0x80;

				for (jj = 0; jj < pixels_per_cell; jj++) {
					fwrite(&pix[0], sizeof(char), 3, fp);
				}
			}

			// Pad end of row to a multiple of four bytes
			remaining_bytes = 4 - ((mxdim * pixels_per_cell * 3) & 3);
			if (remaining_bytes != 4)
				fwrite(&pad[0], sizeof(char), remaining_bytes, fp);
		}
	}

	fclose(fp);
}

// ----------------------------------------------------------------
void store_states_to_bitmap_file(
	int     * current_states,
	int     * next_states,
	int       num_states,
	int       mxdim,
	char    * fn,
	int       pixels_per_cell)
{
	double ** M;
	int i, j;

	M = alloc_matrix(mxdim);

	for (i = 0; i < mxdim; i++)
		for (j = 0; j < mxdim; j++)
			M[i][j] = 0.0;

	for (i = 0; i < num_states; i++)
		M[current_states[i]][next_states[i]] += 1.0;

	store_matrix_to_bitmap_file(M, mxdim, fn, pixels_per_cell);

	free_matrix(M, num_states);
}
